/*------------------------------------------------------------------------------

NAME
      crc -- Compute a cyclic redundancy check on files.

SYNOPSIS
      crc [file]

DESCRIPTION
      A CRC is computed on each input file.  The CRC value (in hex)
	  of each file is printed on standard output, followed by the name
	  of the file.
	  If no argument is given, the CRC of standard input is computed
	  and printed on standard output.
	  
	  The CRC algorithm used in this program is the well known CRC-32.
	  
NO COPYRIGHT
      This is public domain software written by Planar.
	  The CRC lookup table was generated by a program written by Ross Williams.

COMPILATION
    Compile this file with the following commands (with MPW 3.2) :
      C crc.c
      Link -o crc -t MPST -c 'MPS ' crc.c.o 
           "{Libraries}Interface.o" 
	       "{Libraries}Runtime.o" "{CLibraries}StdCLib.o"

------------------------------------------------------------------------------*/

#include <stdio.h>

#include "crc32.h"

unsigned long crcfile (FILE *f){
  CRCaccu crc = CRCinit;
  int c;

  while (1){
    c = getc (f);
	if (c == EOF) break;
	CRCupdate (crc, (unsigned char) c);
  }
  return CRCfinal (crc);
}

int main (int argc, char *argv []){
  FILE *f;
  int i;
  int errcode = 0;
  
  if (argc == 1){
    printf ("%8.8lX\n", crcfile (stdin));
  }else{
    for (i = 1; i < argc; i++){
	  f = fopen (argv [i], "r");
	  if (f == NULL){
	    fprintf (stderr, "crc: cannot open file \"%s\"", argv[i]);
		perror (" ");
		errcode = 3;
	  }else{
	    unsigned long crc = crcfile (f);
		printf ("%8.8lX %s\n", crc, argv [i]);
		fclose (f);
      }
	}
  }
  exit (errcode);
  return 0;  /* not reached */
}
